@@ -5,7 +5,7 @@
     Flask-Security decorators module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -32,6 +32,7 @@ from .utils import (
     check_and_update_authn_fresh,
     json_error_response,
     set_request_attr,
+    get_request_attr,
     url_for_security,
 )
 
@@ -115,7 +116,7 @@ def default_reauthn_handler(within, grace):
     return redirect(redirect_url)
 
 
-def default_unauthz_handler(func, params):
+def default_unauthz_handler(func_name, params):
     unauthz_message, unauthz_message_type = get_message("UNAUTHORIZED")
     if _security._want_json(request):
         payload = json_error_response(errors=unauthz_message)
@@ -141,17 +142,7 @@ def default_unauthz_handler(func, params):
 
 
 def _check_token():
-    # N.B. this isn't great Flask-Login 0.5.0 made this protected
-    # Issue https://github.com/maxcountryman/flask-login/issues/471
-    # was filed to restore public access. We want to call this via
-    # login_manager in case someone has overridden the login_manager which we
-    # allow.
-    if hasattr(_security.login_manager, "request_callback"):
-        # Pre 0.5.0
-        user = _security.login_manager.request_callback(request)
-    else:
-        user = _security.login_manager._request_callback(request)
-
+    user = _security.login_manager.request_callback(request)
     if user and user.is_authenticated:
         app = current_app._get_current_object()
         identity_changed.send(app, identity=Identity(user.fs_uniquifier))
@@ -160,6 +151,25 @@ def _check_token():
     return False
 
 
+def _check_session():
+    """
+    Note that flask_login will have already run _load_user (due to someone referencing
+    current_user proxy). This will have called our _user_loader or _request_loader
+    already.
+    This method needs to determine whether the authenticated user was authenticated
+    due to _user_loader or _request_loader and return True for the former.
+    This routine makes sure that if an endpoint is decorated with just allowing
+    'session' that token authentication won't return True (even though the user
+    is in fact correctly authenticated). There are certainly endpoints that need to
+    be web browser/session only (often those that in fact return tokens!).
+    """
+    if not current_user.is_authenticated:
+        return False
+    if get_request_attr("fs_authn_via") != "session":
+        return False
+    return True
+
+
 def _check_http_auth():
     auth = request.authorization or BasicAuth(username=None, password=None)
     if not auth.username:
@@ -237,8 +247,8 @@ def http_auth_required(realm: t.Any) -> DecoratedView:
                 return _security._unauthorized_callback()
             else:
                 r = _security.default_http_auth_realm if callable(realm) else realm
-                h = {"WWW-Authenticate": 'Basic realm="%s"' % r}
-                return _security._unauthn_handler(["basic"], h)
+                h = {"WWW-Authenticate": f'Basic realm="{r}"'}
+                return _security._unauthn_handler(["basic"], headers=h)
 
         return wrapper
 
@@ -274,7 +284,7 @@ def auth_token_required(fn: DecoratedView) -> DecoratedView:
 def auth_required(
     *auth_methods: t.Union[str, t.Callable[[], t.List[str]], None],
     within: t.Union[int, float, t.Callable[[], datetime.timedelta]] = -1,
-    grace: t.Optional[t.Union[int, float, t.Callable[[], datetime.timedelta]]] = None
+    grace: t.Optional[t.Union[int, float, t.Callable[[], datetime.timedelta]]] = None,
 ) -> DecoratedView:
     """
     Decorator that protects endpoints through multiple mechanisms.
@@ -342,7 +352,7 @@ def auth_required(
 
     login_mechanisms = {
         "token": lambda: _check_token(),
-        "session": lambda: current_user.is_authenticated,
+        "session": lambda: _check_session(),
         "basic": lambda: _check_http_auth(),
     }
     mechanisms_order = ["token", "session", "basic"]
@@ -376,7 +386,7 @@ def auth_required(
             h = {}
             if "basic" in ams:
                 r = _security.default_http_auth_realm
-                h["WWW-Authenticate"] = 'Basic realm="%s"' % r
+                h["WWW-Authenticate"] = f'Basic realm="{r}"'
             mechanisms = [
                 (method, login_mechanisms.get(method))
                 for method in mechanisms_order
@@ -482,7 +492,9 @@ def roles_required(*roles: str) -> DecoratedView:
                     if _security._unauthorized_callback:
                         # Backwards compat - deprecated
                         return _security._unauthorized_callback()
-                    return _security._unauthz_handler(roles_required, list(roles))
+                    return _security._unauthz_handler(
+                        roles_required.__name__, list(roles)
+                    )
             return fn(*args, **kwargs)
 
         return decorated_view
@@ -514,7 +526,7 @@ def roles_accepted(*roles: str) -> DecoratedView:
             if _security._unauthorized_callback:
                 # Backwards compat - deprecated
                 return _security._unauthorized_callback()
-            return _security._unauthz_handler(roles_accepted, list(roles))
+            return _security._unauthz_handler(roles_accepted.__name__, list(roles))
 
         return decorated_view
 
@@ -550,7 +562,7 @@ def permissions_required(*fsperms: str) -> DecoratedView:
                         # Backwards compat - deprecated
                         return _security._unauthorized_callback()
                     return _security._unauthz_handler(
-                        permissions_required, list(fsperms)
+                        permissions_required.__name__, list(fsperms)
                     )
 
             return fn(*args, **kwargs)
@@ -588,7 +600,9 @@ def permissions_accepted(*fsperms: str) -> DecoratedView:
             if _security._unauthorized_callback:
                 # Backwards compat - deprecated
                 return _security._unauthorized_callback()
-            return _security._unauthz_handler(permissions_accepted, list(fsperms))
+            return _security._unauthz_handler(
+                permissions_accepted.__name__, list(fsperms)
+            )
 
         return decorated_view
 
