@@ -1,33 +1,48 @@
-# -*- coding: utf-8 -*-
 """
     flask_security.phone_util
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 
     Utility class for managing phone numbers
 
-    :copyright: (c) 2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2020-2021 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
     Avoid making 'phonenumbers' a required package unless needed.
 """
+import typing as t
 
 from .utils import config_value, get_message
 
+if t.TYPE_CHECKING:  # pragma: no cover
+    import flask
 
-class PhoneUtil(object):
+
+class PhoneUtil:
     """
     Provide parsing and validation for user inputted phone numbers.
     Subclass this to use a different underlying phone number parsing library.
 
     To provide your own implementation, pass in the class as ``phone_util_cls``
-    at init time. Your class will be instantiated once prior to the first
-    request being handled.
+    at init time. Your class will be instantiated once as part of
+    Flask-Security initialization.
 
     .. versionadded:: 3.4.0
+
+    .. versionchanged:: 4.0.0
+        __init__ takes app argument, and is instantiated at Flask-Security
+        initialization time rather than at first request.
     """
 
-    def validate_phone_number(self, input_data):
-        """ Return ``None`` if a valid phone number else an error message. """
+    def __init__(self, app: "flask.Flask"):
+        """Instantiate class.
+
+        :param app: The Flask application being initialized.
+        """
+        pass
+
+    def validate_phone_number(self, input_data: str) -> t.Optional[str]:
+        """Return ``None`` if a valid phone number else
+        the ``PHONE_INVALID`` error message."""
         import phonenumbers
 
         try:
@@ -40,8 +55,8 @@ class PhoneUtil(object):
             pass
         return get_message("PHONE_INVALID")[0]
 
-    def get_canonical_form(self, input_data):
-        """ Validate and return a canonical form to be stored in DB
+    def get_canonical_form(self, input_data: str) -> t.Optional[str]:
+        """Validate and return a canonical form to be stored in DB
         and compared against.
         Returns ``None`` if input isn't a valid phone number.
         """
