@@ -1,4 +1,3 @@
-# -*- coding: utf-8 -*-
 """
     flask_security.signals
     ~~~~~~~~~~~~~~~~~~~~~~
@@ -6,7 +5,7 @@
     Flask-Security signals module
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -18,6 +17,9 @@ user_authenticated = signals.signal("user-authenticated")
 
 user_registered = signals.signal("user-registered")
 
+# For cases of RETURN_GENERIC_RESPONSES with existing email/username
+user_not_registered = signals.signal("user-not-registered")
+
 user_confirmed = signals.signal("user-confirmed")
 
 confirm_instructions_sent = signals.signal("confirm-instructions-sent")
@@ -41,3 +43,7 @@ tf_disabled = signals.signal("tf-disabled")
 us_security_token_sent = signals.signal("us-security-token-sent")
 
 us_profile_changed = signals.signal("us-profile-changed")
+
+wan_registered = signals.signal("wan-registered")
+
+wan_deleted = signals.signal("wan-deleted")
