@@ -5,7 +5,7 @@
     Flask-Security utils module
 
     :copyright: (c) 2012-2019 by Matt Wright.
-    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 import abc
@@ -14,10 +14,9 @@ import datetime
 from functools import partial
 import hashlib
 import hmac
-from pkg_resources import parse_version
 import time
 import typing as t
-from urllib.parse import parse_qsl, parse_qs, urlsplit, urlunsplit, urlencode
+from urllib.parse import parse_qsl, urlsplit, urlunsplit, urlencode
 import urllib.request
 import urllib.error
 import warnings
@@ -37,10 +36,9 @@ from flask_login import logout_user as _logout_user
 from flask_login import current_user
 from flask_login import COOKIE_NAME as REMEMBER_COOKIE_NAME
 from flask_principal import AnonymousIdentity, Identity, identity_changed, Need
-from flask_wtf import csrf
+from flask_wtf import csrf, FlaskForm
 from wtforms import ValidationError
 from itsdangerous import BadSignature, SignatureExpired
-from werkzeug import __version__ as werkzeug_version
 from werkzeug.local import LocalProxy
 from werkzeug.datastructures import MultiDict
 
@@ -51,7 +49,6 @@ from .signals import user_authenticated
 if t.TYPE_CHECKING:  # pragma: no cover
     from flask import Flask, Response
     from flask.typing import ResponseValue
-    from flask_wtf import FlaskForm
     from .datastore import User
 
 SB = t.Union[str, bytes]
@@ -114,7 +111,9 @@ def find_csrf_field_name():
     overridable.
     We take the field name from the login_form as set by the configuration.
     """
-    form = _security.login_form(MultiDict([]))
+    from .forms import DummyForm
+
+    form = DummyForm(formdata=None)
     if hasattr(form.meta, "csrf_field_name"):
         return form.meta.csrf_field_name
     return None
@@ -423,9 +422,7 @@ def do_flash(message: str, category: str) -> None:
         flash(message, category)
 
 
-def get_url(
-    endpoint_or_url: t.Optional[str], qparams: t.Optional[t.Dict[str, str]] = None
-) -> t.Optional[str]:
+def get_url(endpoint_or_url: str, qparams: t.Optional[t.Dict[str, str]] = None) -> str:
     """Returns a URL if a valid endpoint is found. Otherwise, returns the
     provided value.
 
@@ -433,14 +430,13 @@ def get_url(
     :param qparams: additional query params to add to end of url
     :return: URL
     """
-    if not endpoint_or_url:
-        return endpoint_or_url
     try:
         return transform_url(url_for(endpoint_or_url), qparams)
     except Exception:
         # This is an external URL (no endpoint defined in app)
         # For (mostly) testing - allow changing/adding the url - for example
         # add a different host:port for cases where the UI is running
+        # add a different host:port for cases where the UI is running
         # separately.
         if config_value("REDIRECT_HOST"):
             url = transform_url(
@@ -456,13 +452,12 @@ def slash_url_suffix(url, suffix):
     """Adds a slash either to the beginning or the end of a suffix
     (which is to be appended to a URL), depending on whether or not
     the URL ends with a slash."""
-
-    return url.endswith("/") and ("%s/" % suffix) or ("/%s" % suffix)
+    return url.endswith("/") and f"{suffix}/" or f"/{suffix}"
 
 
 def transform_url(
-    url: t.Optional[str], qparams: t.Optional[t.Dict[str, str]] = None, **kwargs: str
-) -> t.Optional[str]:
+    url: str, qparams: t.Optional[t.Dict[str, str]] = None, **kwargs: str
+) -> str:
     """Modify url
 
     :param url: url to transform (can be relative)
@@ -472,8 +467,6 @@ def transform_url(
 
     .. versionadded:: 3.2.0
     """
-    if not url:
-        return url
     link_parse = urlsplit(url)
     if qparams:
         current_query = dict(parse_qsl(link_parse.query))
@@ -486,7 +479,7 @@ def get_security_endpoint_name(endpoint):
     return f"{_security.blueprint_name}.{endpoint}"
 
 
-def url_for_security(endpoint: str, **values: t.Union[str, bool]) -> str:
+def url_for_security(endpoint: str, **values: t.Any) -> str:
     """Return a URL for the security blueprint
 
     :param endpoint: the endpoint of the URL (name of the function)
@@ -554,56 +547,71 @@ def validate_redirect_url(url: str) -> bool:
     return True
 
 
-def get_post_action_redirect(config_key: str, declared: t.Optional[str] = None) -> str:
-    urls = [
-        get_url(request.args.get("next", None)),
-        get_url(request.form.get("next", None)),
-        find_redirect(config_key),
-    ]
-    if declared:
-        urls.insert(0, declared)
-    for url in urls:
-        if url and validate_redirect_url(url):
-            return url
-    raise ValueError("No valid redirect URL found - configuration error")
+def get_post_action_redirect(config_key: str) -> str:
+    return propagate_next(find_redirect(config_key), request.form)
 
 
-def get_post_login_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_LOGIN_VIEW", declared)
+def get_post_login_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_LOGIN_VIEW")
 
 
-def get_post_register_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_REGISTER_VIEW", declared)
+def get_post_register_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_REGISTER_VIEW")
 
 
-def get_post_logout_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_LOGOUT_VIEW", declared)
+def get_post_logout_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_LOGOUT_VIEW")
 
 
-def get_post_verify_redirect(declared: t.Optional[str] = None) -> str:
-    return get_post_action_redirect("SECURITY_POST_VERIFY_VIEW", declared)
+def get_post_verify_redirect() -> str:
+    return get_post_action_redirect("SECURITY_POST_VERIFY_VIEW")
 
 
-def find_redirect(key: str) -> t.Optional[str]:
+def find_redirect(key: str) -> str:
     """Returns the URL to redirect to after a user logs in successfully.
 
     :param key: The session or application configuration key to search for
     """
-    rv = (
-        get_url(session.pop(key.lower(), None))
-        or get_url(current_app.config[key.upper()] or None)
-        or current_app.config.get("APPLICATION_ROOT", "/")
-    )
+    session_value = session.pop(key.lower(), None)
+    session_url = None
+    if session_value:
+        session_url = get_url(session_value)
+    app_value = current_app.config[key.upper()]
+    app_url = None
+    if app_value:
+        app_url = get_url(app_value)
+    rv = session_url or app_url or str(current_app.config.get("APPLICATION_ROOT", "/"))
     return rv
 
 
-def propagate_next(url: str) -> str:
-    # return either URL or, if URL already has a ?next=xx, return that.
-    url_next = urlsplit(url)
-    qparams = parse_qs(url_next.query)
-    if "next" in qparams:
-        return qparams["next"][0]
-    return url
+def propagate_next(
+    fallback_url: str, form: t.Optional[t.Union[FlaskForm, MultiDict]]
+) -> str:
+    """Compute appropriate redirect URL
+    The application can add a 'next' query parameter or have 'next' as a form field.
+    If either exist, make sure they are valid (not pointing to external location)
+    If neither, return the fallback_url
+
+    Can be passed either request.form (which is really a MultiDict OR a real form.
+    """
+    form_next = None
+    if form and isinstance(form, FlaskForm):
+        if hasattr(form, "next") and form.next.data:
+            form_next = get_url(form.next.data)
+    elif form and form.get("next", None):
+        form_next = get_url(str(form.get("next")))
+    arg_next = None
+    if request.args.get("next", None):
+        arg_next = get_url(request.args.get("next"))  # type: ignore[arg-type]
+    urls = [
+        arg_next,
+        form_next,
+        fallback_url,
+    ]
+    for url in urls:
+        if url and validate_redirect_url(url):
+            return url
+    raise ValueError("No valid redirect URL found - configuration error")
 
 
 def get_config(app: "Flask") -> t.Dict[str, t.Any]:
@@ -685,11 +693,11 @@ def send_mail(subject, recipient, template, **context):
 
     body = None
     html = None
-    ctx = ("security/email", template)
+    template_path = f"security/email/{template}"
     if config_value("EMAIL_PLAINTEXT"):
-        body = _security.render_template("%s/%s.txt" % ctx, **context)
+        body = _security.render_template(f"{template_path}.txt", **context)
     if config_value("EMAIL_HTML"):
-        html = _security.render_template("%s/%s.html" % ctx, **context)
+        html = _security.render_template(f"{template_path}.html", **context)
 
     subject = localize_callback(subject)
 
@@ -698,7 +706,13 @@ def send_mail(subject, recipient, template, **context):
         sender = sender._get_current_object()
 
     _security._mail_util.send_mail(
-        template, subject, recipient, sender, body, html, context.get("user", None)
+        template,
+        subject,
+        recipient,
+        sender,
+        body,
+        html,
+        **context,
     )
 
 
@@ -717,6 +731,7 @@ def get_token_status(token, serializer, max_age=None, return_data=False):
     warnings.warn(
         "'get_token_status' is deprecated - use check_and_get_token_status instead",
         DeprecationWarning,
+        stacklevel=2,
     )
     serializer = getattr(_security, serializer + "_serializer")
     max_age = get_max_age(max_age)
@@ -911,10 +926,7 @@ def csrf_cookie_handler(response: "Response") -> "Response":
 
     if op == "clear":
         # Alas delete_cookie only accepts some of the keywords set_cookie does
-        # and Werkzeug didn't accept samesite, secure, httponly until 2.0
         allowed = ["path", "domain", "secure", "httponly", "samesite"]
-        if parse_version(werkzeug_version) < parse_version("2.0.0"):  # pragma: no cover
-            allowed = ["path", "domain"]
         args = {k: csrf_cookie.get(k) for k in allowed if k in csrf_cookie}
         response.delete_cookie(csrf_cookie_name, **args)
         session.pop("fs_cc")
@@ -1031,8 +1043,8 @@ def json_error_response(
     """Helper to create an error response.
 
     The "errors" key holds a simple list of errors - which is made up of any passed
-    errors (either a string or list) as well as the (localized) error msg from the
-    passed in form errors.
+    errors (either a string or list) as well as the (localized) error msgs from the
+    passed in field_errors.
 
     The "field_errors" key which is exactly what is returned from WTForms - namely
     a dict of field-name: msg. For form-level errors (WTForms 3.0) the 'field-name' is
@@ -1063,9 +1075,6 @@ def default_render_template(*args, **kwargs):
 
 
 class SmsSenderBaseClass(metaclass=abc.ABCMeta):
-    def __init__(self, *args, **kwargs):
-        pass
-
     @abc.abstractmethod
     def send_sms(
         self, from_number: str, to_number: str, msg: str
