@@ -4,23 +4,27 @@
 
     Test common functionality
 
-    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
 import base64
 import json
 import re
-from http.cookiejar import Cookie
 import pytest
 
-from flask import Blueprint
+from flask import Blueprint, g
 
 from flask_security import uia_email_mapper
+from flask_security.decorators import auth_required
+from flask_principal import identity_loaded
 
 from tests.test_utils import (
     authenticate,
+    capture_flashes,
+    check_location,
     get_auth_token_version_3x,
+    get_form_action,
     get_num_queries,
     hash_password,
     json_authenticate,
@@ -74,6 +78,19 @@ def test_authenticate_with_invalid_malformed_next(client, get_message):
     assert get_message("INVALID_REDIRECT") in response.data
 
 
+def test_login_template_next(client):
+    # Test that our login template propagates next.
+    response = client.get("/profile", follow_redirects=True)
+    assert "?next=/profile" in response.request.url
+    login_url = get_form_action(response)
+    response = client.post(
+        login_url,
+        data=dict(email="matt@lp.com", password="password"),
+        follow_redirects=True,
+    )
+    assert b"Profile Page" in response.data
+
+
 def test_authenticate_with_subdomain_next(app, client, get_message):
     app.config["SERVER_NAME"] = "lp.com"
     app.config["SECURITY_REDIRECT_ALLOW_SUBDOMAINS"] = True
@@ -166,6 +183,16 @@ def test_login_form_username(client):
     assert re.search(b'<input[^>]*autocomplete="current-password"[^>]*>', response.data)
 
 
+@pytest.mark.settings(username_enable=True, username_required=True)
+def test_login_form_username_required(app, client):
+    # If username required - we should still be able to login with email alone
+    # given default user_identity_attributes
+    response = client.post(
+        "/login", data=dict(email="matt@lp.com", password="password")
+    )
+    assert check_location(app, response.location, "/")
+
+
 @pytest.mark.confirmable()
 @pytest.mark.settings(
     return_generic_responses=True, requires_confirmation_error_view="/confirm"
@@ -193,12 +220,14 @@ def test_generic_response(app, client, get_message):
     )
 
     # make sure don't get confirmation required
-    response = client.post(
-        "/login",
-        data=dict(email="mattwho@lp.com", password="password"),
-        follow_redirects=False,
-    )
-    assert response.status_code == 200
+    with capture_flashes() as flashes:
+        response = client.post(
+            "/login",
+            data=dict(email="mattwho@lp.com", password="password"),
+            follow_redirects=False,
+        )
+        assert response.status_code == 200
+    assert len(flashes) == 0
 
 
 @pytest.mark.registerable()
@@ -504,6 +533,17 @@ def test_token_auth_via_header_invalid_token(client):
     assert response.status_code == 401
 
 
+def test_token_auth_invalid_for_session_auth(client):
+    # when user is loaded from token data, session authentication should fail.
+    response = json_authenticate(client)
+    token = response.json["response"]["user"]["authentication_token"]
+    # logout so session doesn't contain valid user details
+    logout(client)
+    headers = {"Authentication-Token": token, "Accept": "application/json"}
+    response = client.get("/session", headers=headers)
+    assert response.status_code == 401
+
+
 def test_http_auth(client):
     # browsers expect 401 response with WWW-Authenticate header - which will prompt
     # them to pop up a login form.
@@ -718,35 +758,37 @@ def test_user_deleted_during_session_reverts_to_anonymous_user(app, client):
     assert b"Hello matt@lp.com" not in response.data
 
 
+def test_session_loads_identity(app, client):
+    @app.route("/identity_check")
+    @auth_required("session")
+    def id_check():
+        if hasattr(g, "identity"):
+            identity = g.identity
+            assert hasattr(identity, "loader_called")
+            assert identity.loader_called
+        return "Success"
+
+    json_authenticate(client)
+
+    # add identity loader after authentication to only fire it for
+    # session-authentication next `get` call
+    @identity_loaded.connect_via(app)
+    def identity_loaded_check(sender, identity):
+        identity.loader_called = True
+
+    response = client.get("/identity_check")
+    assert b"Success" == response.data
+
+
 def test_remember_token(client):
     response = authenticate(client, follow_redirects=False)
-    client.cookie_jar.clear_session_cookies()
+    client.delete_cookie("session")
     response = client.get("/profile")
     assert b"profile" in response.data
 
 
 def test_request_loader_does_not_fail_with_invalid_token(client):
-    c = Cookie(
-        version=0,
-        name="remember_token",
-        value="None",
-        port=None,
-        port_specified=False,
-        domain="www.example.com",
-        domain_specified=False,
-        domain_initial_dot=False,
-        path="/",
-        path_specified=True,
-        secure=False,
-        expires=None,
-        discard=True,
-        comment=None,
-        comment_url=None,
-        rest={"HttpOnly": None},
-        rfc2109=False,
-    )
-
-    client.cookie_jar.set_cookie(c)
+    client.set_cookie("remember_token")
     response = client.get("/")
     assert b"BadSignature" not in response.data
 
@@ -1022,7 +1064,6 @@ def test_auth_token_decorator(in_app_context):
     client_nc = app.test_client(use_cookies=False)
 
     with app.test_request_context("/"):
-
         user = app.security.datastore.find_user(email="matt@lp.com")
         token = get_auth_token_version_3x(app, user)
 
