@@ -102,10 +102,9 @@ def test_context_processors(client, app):
 
     client.get("/logout")
 
-    with app.mail.record_messages() as outbox:
-        client.post("/reset", data=dict(email="matt@lp.com"))
+    client.post("/reset", data=dict(email="matt@lp.com"))
 
-    email = outbox[0]
+    email = app.mail.outbox[1]
     assert "global" in email.body
     assert "bar-mail" in email.body
 
@@ -177,7 +176,7 @@ def test_unified_signin_context_processors(client, app):
     assert b"global" in response.data
     assert b"signin" in response.data
 
-    response = client.get("/us-signin/send-code")
+    response = client.post("/us-signin/send-code")
     assert b"CUSTOM UNIFIED SIGN IN" in response.data
     assert b"global" in response.data
     assert b"signin" in response.data
@@ -213,3 +212,34 @@ def test_unified_signin_context_processors(client, app):
     assert b"CUSTOM UNIFIED VERIFY" in response.data
     assert b"global" in response.data
     assert b"setup" in response.data
+
+
+@pytest.mark.two_factor()
+@pytest.mark.settings(
+    multi_factor_recovery_template="custom_security/mf_recovery.html",
+    multi_factor_recovery_codes_template="custom_security/mf_recovery_codes.html",
+    multi_factor_recovery_codes=True,
+)
+def test_mf_recovery_context_processors(client, app):
+    @app.security.context_processor
+    def default_ctx_processor():
+        return {"global": "global"}
+
+    @app.security.mf_recovery_codes_context_processor
+    def codes_ctx():
+        return {"foo": "codes"}
+
+    authenticate(client)
+    response = client.get("/mf-recovery-codes")
+    assert b"global" in response.data
+    assert b"codes" in response.data
+    logout(client)
+
+    @app.security.mf_recovery_context_processor
+    def code_ctx():
+        return {"foo": "code"}
+
+    authenticate(client, "gal@lp.com")
+    response = client.get("/mf-recovery")
+    assert b"global" in response.data
+    assert b"code" in response.data
