@@ -4,7 +4,7 @@
 
     CSRF tests
 
-    :copyright: (c) 2019-2020 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 from contextlib import contextmanager
@@ -15,6 +15,7 @@ import flask_wtf.csrf
 import pytest
 from flask_wtf import CSRFProtect
 
+from flask_security import Security, hash_password
 from tests.test_utils import get_session, logout
 
 
@@ -60,6 +61,15 @@ def _get_csrf_token(client):
     return response.json["response"]["csrf_token"]
 
 
+def create_user(app):
+    with app.app_context():
+        app.security.datastore.create_user(
+            email="matt@lp.com",
+            password=hash_password("password"),
+        )
+        app.security.datastore.commit()
+
+
 def json_login(
     client,
     email="matt@lp.com",
@@ -156,9 +166,13 @@ def test_login_csrf_json(app, client):
     assert "csrf_token" not in session
 
 
-def test_login_csrf_json_header(app, client):
+def test_login_csrf_json_header(app, sqlalchemy_datastore):
     app.config["WTF_CSRF_ENABLED"] = True
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
+
+    client = app.test_client()
 
     with mp_validate_csrf() as mp:
         auth_token, csrf_token = json_login(client, use_header=True)
@@ -196,7 +210,7 @@ def test_login_csrf_unauth_double(app, client, get_message):
     # login in again - should work
     response = client.post("/login", content_type="application/json", json=data)
     assert response.status_code == 400
-    assert response.json["response"]["error"].encode("utf-8") == get_message(
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
         "ANONYMOUS_USER_REQUIRED"
     )
 
@@ -247,10 +261,14 @@ def test_cp_reset(app, client):
 
 
 @pytest.mark.changeable()
-def test_cp_with_token(app, client):
+def test_cp_with_token(app, sqlalchemy_datastore):
     # Make sure can use returned CSRF-Token in Header.
     app.config["WTF_CSRF_ENABLED"] = True
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
+
+    client = app.test_client()
 
     auth_token, csrf_token = json_login(client, use_header=True)
 
@@ -273,10 +291,13 @@ def test_cp_with_token(app, client):
     json_logout(client)
 
 
-def test_cp_login_json_no_session(app, client_nc):
+def test_cp_login_json_no_session(app, sqlalchemy_datastore):
     # Test with global CSRFProtect on and not sending cookie - nothing works.
     app.config["WTF_CSRF_ENABLED"] = True
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+
+    client_nc = app.test_client(use_cookies=False)
 
     # This shouldn't log in - and will return 400
     with mp_validate_csrf() as mp:
@@ -303,38 +324,44 @@ def test_cp_login_json_no_session(app, client_nc):
 
 
 @pytest.mark.settings(CSRF_PROTECT_MECHANISMS=["basic", "session"])
-def test_cp_config(app, client):
+def test_cp_config(app, sqlalchemy_datastore):
     # Test improper config (must have WTF_CSRF_CHECK_DEFAULT false if setting
     # CSRF_PROTECT_MECHANISMS
+    from flask_security import Security
+
     app.config["WTF_CSRF_ENABLED"] = True
     CSRFProtect(app)
 
     # The check is done on first request.
     with pytest.raises(ValueError) as ev:
-        logout(client)
+        Security(app=app, datastore=sqlalchemy_datastore)
     assert "must be set to False" in str(ev.value)
 
 
 @pytest.mark.settings(CSRF_PROTECT_MECHANISMS=["basic", "session"])
-def test_cp_config2(app, client):
+def test_cp_config2(app, sqlalchemy_datastore):
     # Test improper config (must have CSRFProtect configured if setting
     # CSRF_PROTECT_MECHANISMS
+    from flask_security import Security
+
     app.config["WTF_CSRF_ENABLED"] = True
 
-    # The check is done on first request.
     with pytest.raises(ValueError) as ev:
-        logout(client)
+        Security(app=app, datastore=sqlalchemy_datastore)
     assert "CsrfProtect not part of application" in str(ev.value)
 
 
 @pytest.mark.changeable()
 @pytest.mark.settings(CSRF_PROTECT_MECHANISMS=["basic", "session"])
-def test_different_mechanisms(app, client):
+def test_different_mechanisms(app, sqlalchemy_datastore):
     # Verify that using token doesn't require CSRF, but sessions do
     app.config["WTF_CSRF_ENABLED"] = True
     app.config["WTF_CSRF_CHECK_DEFAULT"] = False
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
 
+    client = app.test_client()
     with mp_validate_csrf() as mp:
         auth_token, csrf_token = json_login(client)
 
@@ -398,33 +425,33 @@ def test_different_mechanisms_nc(app, client_nc):
 
 
 @pytest.mark.settings(csrf_ignore_unauth_endpoints=True, CSRF_COOKIE_NAME="XSRF-Token")
-def test_csrf_cookie(app, client):
+def test_csrf_cookie(app, sqlalchemy_datastore):
     app.config["WTF_CSRF_ENABLED"] = True
     app.config["WTF_CSRF_CHECK_DEFAULT"] = False
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
 
+    client = app.test_client()
     json_login(client)
-    found = False
-    for cookie in client.cookie_jar:
-        if cookie.name == "XSRF-Token":
-            found = True
-            assert cookie.path == "/"
-            # Alas http.cookiejar doesn't support samesite...
-            # assert cookie.samesite == "Strict"
-    assert found
+    assert client.get_cookie("XSRF-Token")
 
     # Make sure cleared on logout
     response = client.post("/logout", content_type="application/json")
     assert response.status_code == 200
-    assert "XSRF-Token" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("XSRF-Token")
 
 
 @pytest.mark.settings(CSRF_COOKIE={"key": "XSRF-Token"})
 @pytest.mark.changeable()
-def test_cp_with_token_cookie(app, client):
+def test_cp_with_token_cookie(app, sqlalchemy_datastore):
     # Make sure can use returned CSRF-Token cookie in Header when changing password.
     app.config["WTF_CSRF_ENABLED"] = True
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
+
+    client = app.test_client()
 
     json_login(client, use_header=True)
 
@@ -434,28 +461,32 @@ def test_cp_with_token_cookie(app, client):
         new_password="battery staple",
         new_password_confirm="battery staple",
     )
-    csrf_token = [c.value for c in client.cookie_jar if c.name == "XSRF-Token"][0]
+    csrf_token = client.get_cookie("XSRF-Token")
     with mp_validate_csrf() as mp:
         response = client.post(
             "/change",
             content_type="application/json",
             json=data,
-            headers={"X-XSRF-Token": csrf_token},
+            headers={"X-XSRF-Token": csrf_token.value},
         )
         assert response.status_code == 200
     assert mp.success == 1 and mp.failure == 0
     json_logout(client)
-    assert "XSRF-Token" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("XSRF-Token")
 
 
 @pytest.mark.settings(CSRF_COOKIE_NAME="XSRF-Token", csrf_ignore_unauth_endpoints=True)
 @pytest.mark.changeable()
-def test_cp_with_token_cookie_expire(app, client):
+def test_cp_with_token_cookie_expire(app, sqlalchemy_datastore):
     # Make sure that we get a new Csrf-Token cookie if expired.
     app.config["WTF_CSRF_ENABLED"] = True
     app.config["WTF_CSRF_TIME_LIMIT"] = 1
     app.config["WTF_CSRF_CHECK_DEFAULT"] = False
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
+
+    client = app.test_client()
 
     json_login(client, use_header=True)
 
@@ -466,36 +497,38 @@ def test_cp_with_token_cookie_expire(app, client):
         new_password="battery staple",
         new_password_confirm="battery staple",
     )
-    csrf_token = [c.value for c in client.cookie_jar if c.name == "XSRF-Token"][0]
+    csrf_token = client.get_cookie("XSRF-Token")
     with mp_validate_csrf() as mp:
         response = client.post(
             "/change",
             content_type="application/json",
             json=data,
-            headers={"X-XSRF-Token": csrf_token},
+            headers={"X-XSRF-Token": csrf_token.value},
         )
         assert response.status_code == 400
         assert b"expired" in response.data
 
         # Should have gotten a new CSRF cookie value
-        new_csrf_token = [c.value for c in client.cookie_jar if c.name == "XSRF-Token"][
-            0
-        ]
-        assert csrf_token != new_csrf_token
+        new_csrf_token = client.get_cookie("XSRF-Token")
+        assert csrf_token.value != new_csrf_token.value
     # 2 failures since the utils:csrf_cookie_handler will check
     assert mp.success == 0 and mp.failure == 2
     json_logout(client)
-    assert "XSRF-Token" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("XSRF-Token")
 
 
 @pytest.mark.settings(
     CSRF_COOKIE_NAME="XSRF-Token", CSRF_COOKIE_REFRESH_EACH_REQUEST=True
 )
 @pytest.mark.changeable()
-def test_cp_with_token_cookie_refresh(app, client):
+def test_cp_with_token_cookie_refresh(app, sqlalchemy_datastore):
     # Test CSRF_COOKIE_REFRESH_EACH_REQUEST
     app.config["WTF_CSRF_ENABLED"] = True
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
+
+    client = app.test_client()
 
     json_login(client, use_header=True)
 
@@ -506,10 +539,10 @@ def test_cp_with_token_cookie_refresh(app, client):
         new_password_confirm="battery staple",
     )
 
-    csrf_cookie = [c for c in client.cookie_jar if c.name == "XSRF-Token"][0]
+    csrf_cookie = client.get_cookie("XSRF-Token")
     with mp_validate_csrf() as mp:
         # Delete cookie - we should always get a new one
-        client.delete_cookie(csrf_cookie.domain, csrf_cookie.name)
+        client.delete_cookie("XSRF-Token")
         response = client.post(
             "/change",
             content_type="application/json",
@@ -517,46 +550,50 @@ def test_cp_with_token_cookie_refresh(app, client):
             headers={"X-XSRF-Token": csrf_cookie.value},
         )
         assert response.status_code == 200
-        csrf_cookie = [c for c in client.cookie_jar if c.name == "XSRF-Token"][0]
-        assert csrf_cookie
+        assert client.get_cookie("XSRF-Token")
     assert mp.success == 1 and mp.failure == 0
 
     # delete cookie again, do a 'GET' - the REFRESH_COOKIE_ON_EACH_REQUEST should
     # send us a new one
-    client.delete_cookie(csrf_cookie.domain, csrf_cookie.name)
+    client.delete_cookie("XSRF-Token")
     response = client.get("/change")
     assert response.status_code == 200
-    csrf_cookie = [c for c in client.cookie_jar if c.name == "XSRF-Token"][0]
-    assert csrf_cookie
+    assert client.get_cookie("XSRF-Token")
 
     json_logout(client)
-    assert "XSRF-Token" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("XSRF-Token")
 
 
 @pytest.mark.settings(CSRF_COOKIE_NAME="XSRF-Token")
 @pytest.mark.changeable()
-def test_remember_login_csrf_cookie(app, client):
+def test_remember_login_csrf_cookie(app, sqlalchemy_datastore):
     # Test csrf cookie upon resuming a remember session
     app.config["WTF_CSRF_ENABLED"] = True
     CSRFProtect(app)
+    app.security = Security(app=app, datastore=sqlalchemy_datastore)
+    create_user(app)
+
+    client = app.test_client()
 
     # Login with remember_token generation
     json_login(client, use_header=True, remember=True)
 
-    csrf_cookie = [c for c in client.cookie_jar if c.name == "XSRF-Token"][0]
-    session_cookie = [c for c in client.cookie_jar if c.name == "session"][0]
+    # csrf_cookie = [c for c in client.cookie_jar if c.name == "XSRF-Token"][0]
+    # session_cookie = [c for c in client.cookie_jar if c.name == "session"][0]
     # Delete session and csrf cookie - we should always get new ones
-    client.delete_cookie(csrf_cookie.domain, csrf_cookie.name)
-    client.delete_cookie(session_cookie.domain, session_cookie.name)
+    # client.delete_cookie(csrf_cookie.domain, csrf_cookie.name)
+    # client.delete_cookie(session_cookie.domain, session_cookie.name)
+    client.delete_cookie("XSRF-Token")
+    client.delete_cookie("session")
 
     # Do a simple get request with the remember_token cookie present
-    assert "remember_token" in [c.name for c in client.cookie_jar]
+    assert client.get_cookie("remember_token")
     response = client.get("/profile")
     assert response.status_code == 200
-    assert "session" in [c.name for c in client.cookie_jar]
-    assert "XSRF-Token" in [c.name for c in client.cookie_jar]
+    assert client.get_cookie("session")
+    assert client.get_cookie("XSRF-Token")
     # Logout and check that everything cleans up nicely
     json_logout(client)
-    assert "remember_token" not in [c.name for c in client.cookie_jar]
-    assert "session" not in [c.name for c in client.cookie_jar]
-    assert "XSRF-Token" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("remember_token")
+    assert not client.get_cookie("session")
+    assert not client.get_cookie("XSRF-Token")
