@@ -5,7 +5,7 @@
     Lots of tests
 
     :copyright: (c) 2012 by Matt Wright.
-    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 
@@ -14,8 +14,6 @@ import hashlib
 from unittest import mock
 import re
 import os.path
-import pkg_resources
-import sys
 import time
 import typing as t
 
@@ -27,7 +25,9 @@ from tests.test_utils import (
     authenticate,
     capture_flashes,
     capture_reset_password_requests,
+    check_location,
     check_xlation,
+    get_csrf_token,
     init_app_with_options,
     json_authenticate,
     logout,
@@ -42,6 +42,7 @@ from flask_security.forms import (
     ChangePasswordForm,
     ConfirmRegisterForm,
     EmailField,
+    EmailValidation,
     ForgotPasswordForm,
     LoginForm,
     PasswordField,
@@ -52,7 +53,6 @@ from flask_security.forms import (
     SendConfirmationForm,
     StringField,
     email_required,
-    email_validator,
     valid_user_email,
 )
 from flask_security import auth_required, roles_required
@@ -124,7 +124,7 @@ def test_basic_custom_forms(app, sqlalchemy_datastore):
     class MyForgotPasswordForm(ForgotPasswordForm):
         email = EmailField(
             "My Forgot Email Address Field",
-            validators=[email_required, email_validator, valid_user_email],
+            validators=[email_required, EmailValidation(verify=True), valid_user_email],
         )
 
     class MyResetPasswordForm(ResetPasswordForm):
@@ -266,7 +266,7 @@ def test_passwordless_and_two_factor_configuration_mismatch(app, sqlalchemy_data
         init_app_with_options(
             app,
             sqlalchemy_datastore,
-            **{"SECURITY_TWO_FACTOR": True, "SECURITY_TWO_FACTOR_ENABLED_METHODS": []}
+            **{"SECURITY_TWO_FACTOR": True, "SECURITY_TWO_FACTOR_ENABLED_METHODS": []},
         )
 
 
@@ -294,7 +294,7 @@ def test_change_hash_type(app, sqlalchemy_datastore):
             "SECURITY_PASSWORD_HASH": "plaintext",
             "SECURITY_PASSWORD_SALT": None,
             "SECURITY_PASSWORD_SCHEMES": ["bcrypt", "plaintext"],
-        }
+        },
     )
 
     app.config["SECURITY_PASSWORD_HASH"] = "bcrypt"
@@ -396,6 +396,109 @@ def test_custom_forms_via_config(app, sqlalchemy_datastore):
     assert b"My Register Email Address Field" in response.data
 
 
+def test_custom_form_instantiator(app, client, get_message):
+    # Test application form instantiation.
+    # This is using the form factory pattern.
+    # Note in this case - Flask-Security doesn't even know the form class name.
+    from flask_security import FormInfo
+
+    class FormInstantiator:
+        def __init__(self, myservice):
+            self.myservice = myservice
+
+        def instantiator(self, form_name, form_cls, *args, **kwargs):
+            if form_name == "login_form":
+                return MyLoginForm(*args, service=self.myservice, **kwargs)
+            raise ValueError("Unknown Form")
+
+    class MyLoginForm(LoginForm):
+        def __init__(self, *args, service=None, **kwargs):
+            super().__init__(*args, **kwargs)
+            self.myservice = service
+
+        def validate(self, **kwargs: t.Any) -> bool:
+            if not super().validate(**kwargs):  # pragma: no cover
+                return False
+            if not self.myservice(self.email.data):
+                self.email.errors.append("Not happening")
+                return False
+            return True
+
+    def login_checker(email):
+        return True if email == "matt@lp.com" else False
+
+    fi = FormInstantiator(login_checker)
+    app.security.set_form_info("login_form", FormInfo(fi.instantiator))
+
+    response = authenticate(client, follow_redirects=True)
+    assert b"Welcome matt@lp.com" in response.data
+    logout(client)
+
+    # Try a normally legit user - but our service denies it
+    response = authenticate(client, email="joe@lp.com")
+    assert b"Not happening" in response.data
+
+
+def test_custom_form_instantiator2(app, client, get_message):
+    # Test application form instantiation.
+    # This is using the form clone pattern.
+    # Note in this case - Flask-Security doesn't even know the form class name.
+    app.config["WTF_CSRF_ENABLED"] = True
+    from flask_security import FormInfo
+
+    class MyLoginForm(LoginForm):
+        def __init__(self, *args, service=None, **kwargs):
+            super().__init__(*args, **kwargs)
+            self.myservice = service
+
+        def instantiator(self, form_name, form_cls, *args, **kwargs):
+            return MyLoginForm(*args, service=self.myservice, **kwargs)
+
+        def validate(self, **kwargs: t.Any) -> bool:
+            if not super().validate(**kwargs):  # pragma: no cover
+                return False
+            if not self.myservice(self.email.data):
+                self.email.errors.append("Not happening")
+                return False
+            return True
+
+    def login_checker(email):
+        return True if email == "matt@lp.com" else False
+
+    with app.test_request_context():
+        fi = MyLoginForm(formdata=None, service=login_checker)
+    app.security.set_form_info("login_form", FormInfo(fi.instantiator))
+
+    csrf_token = get_csrf_token(client)
+    response = client.post(
+        "/login",
+        data=dict(email="matt@lp.com", password="password", csrf_token=csrf_token),
+        follow_redirects=True,
+    )
+    assert b"Welcome matt@lp.com" in response.data
+    logout(client)
+
+    # Try a normally legit user - but our service denies it
+    csrf_token = get_csrf_token(client)
+    response = client.post(
+        "/login",
+        data=dict(email="joe@lp.com", password="password", csrf_token=csrf_token),
+    )
+    assert b"Not happening" in response.data
+
+
+def test_custom_form_setting(app, sqlalchemy_datastore):
+    from flask_security import FormInfo
+
+    security = Security(app=app, datastore=sqlalchemy_datastore)
+    with pytest.raises(ValueError) as vex:
+        security.set_form_info("mylogin_form", FormInfo())
+    assert "Unknown form name mylogin_form" == str(vex.value)
+    with pytest.raises(ValueError) as vex:
+        security.set_form_info("login_form", FormInfo())
+    assert "form class must be provided" in str(vex.value)
+
+
 def test_form_required(app, sqlalchemy_datastore):
     class MyLoginForm(LoginForm):
         myfield = StringField("My Custom Field", validators=[Required()])
@@ -487,9 +590,35 @@ def test_sender_tuple(app, sqlalchemy_datastore):
         assert outbox[0].from_email == "Test User <test@testme.com>"
 
 
+def test_send_mail_context(app, sqlalchemy_datastore):
+    """Test full context sent to MailUtil/send_mail"""
+    app.config["MAIL_DEFAULT_SENDER"] = "test@testme.com"
+    app.security = Security()
+    app.security.init_app(app, sqlalchemy_datastore)
+
+    class TestUser:
+        def __init__(self, email):
+            self.email = email
+
+    @app.security.mail_context_processor
+    def mail():
+        return {"foo": "bar-mail"}
+
+    with app.app_context():
+        user = TestUser("matt@lp.com")
+        send_mail("Test Default Sender", user.email, "welcome", user=user)
+        outbox = app.mail.outbox
+        assert 1 == len(outbox)
+        assert "test@testme.com" == outbox[0].from_email
+        matcher = re.match(
+            r".*ExtraContext:(\S+).*", outbox[0].body, re.IGNORECASE | re.DOTALL
+        )
+        assert matcher.group(1) == "bar-mail"
+
+
 @pytest.mark.babel()
+@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_xlation(app, client):
-    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
 
     response = client.get("/login")
@@ -501,25 +630,20 @@ def test_xlation(app, client):
 
 
 @pytest.mark.babel()
+@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
     # Test changing a single MSG and having an additional translation dir
     # Flask-BabelEx doesn't support lists of directories..
-    try:
-        import flask_babelex  # noqa: F401
-
-        pytest.skip("Flask-BabelEx doesn't support lists of translations")
-    except ImportError:
-        pass
+    pytest.importorskip("flask_babel")
 
     i18n_dirname = [
-        pkg_resources.resource_filename("flask_security", "translations"),
+        "builtin",
         os.path.join(pytestconfig.rootdir, "tests/translations"),
     ]
     init_app_with_options(
         app, sqlalchemy_datastore, **{"SECURITY_I18N_DIRNAME": i18n_dirname}
     )
 
-    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
 
     app.config["SECURITY_MSG_INVALID_PASSWORD"] = ("Password no-worky", "error")
@@ -530,8 +654,29 @@ def test_myxlation(app, sqlalchemy_datastore, pytestconfig):
 
 
 @pytest.mark.babel()
+@pytest.mark.app_settings(babel_default_locale="fr_FR")
+def test_myxlation_complete(app, sqlalchemy_datastore, pytestconfig):
+    # Test having own translations and not using builtin.
+    pytest.importorskip("flask_babel")
+    i18n_dirname = [
+        os.path.join(pytestconfig.rootdir, "tests/translations"),
+    ]
+    init_app_with_options(
+        app, sqlalchemy_datastore, **{"SECURITY_I18N_DIRNAME": i18n_dirname}
+    )
+
+    assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
+
+    app.config["SECURITY_MSG_INVALID_PASSWORD"] = ("Password no-worky", "error")
+
+    client = app.test_client()
+    response = client.post("/login", data=dict(email="matt@lp.com", password="forgot"))
+    assert b"Passe - no-worky" in response.data
+
+
+@pytest.mark.babel()
+@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_form_labels(app, sqlalchemy_datastore):
-    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     app.security = Security()
     app.security.init_app(app, sqlalchemy_datastore)
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
@@ -557,6 +702,7 @@ def test_form_labels(app, sqlalchemy_datastore):
 
 
 @pytest.mark.babel()
+@pytest.mark.app_settings(babel_default_locale="fr_FR")
 def test_wtform_xlation(app, sqlalchemy_datastore):
     # Make sure wtform xlations work
     class MyLoginForm(LoginForm):
@@ -564,7 +710,6 @@ def test_wtform_xlation(app, sqlalchemy_datastore):
             "FixedLength", validators=[DataRequired(), Length(3, 3)]
         )
 
-    app.config["BABEL_DEFAULT_LOCALE"] = "fr_FR"
     app.security = Security()
     app.security.init_app(app, datastore=sqlalchemy_datastore, login_form=MyLoginForm)
     assert check_xlation(app, "fr_FR"), "You must run python setup.py compile_catalog"
@@ -594,7 +739,6 @@ def test_per_request_xlate(app, client):
 
     babel = app.extensions["babel"]
 
-    @babel.localeselector
     def get_locale():
         # For a given session - set lang based on first request.
         # Honor explicit url request first
@@ -606,6 +750,9 @@ def test_per_request_xlate(app, client):
                 session["lang"] = locale
         return session.get("lang", None).replace("-", "_")
 
+    babel.locale_selector_func = get_locale
+    babel.locale_selector = get_locale  # Flask-Babel >= 3.0.0
+
     response = client.get("/login", headers=[("Accept-Language", "fr")])
     assert b'<label for="password">Mot de passe</label>' in response.data
     # make sure template contents get xlated (not just form).
@@ -662,10 +809,8 @@ def test_zxcvbn_xlate(app):
 """
 
 
-@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
 @pytest.mark.settings(password_check_breached="strict")
 def test_breached(app, sqlalchemy_datastore):
-
     # partial response from: https://api.pwnedpasswords.com/range/07003
     pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
 AFD8AA47E6FD782ADDC11D89744769F7354:2\r\n\
@@ -687,14 +832,12 @@ B3902FD808DCA504AAAD30F3C14BD3ACE7C:10"
             assert app.config["SECURITY_MSG_PASSWORD_BREACHED"][0] in pbad[0]
 
 
-@pytest.mark.skipif(sys.version_info < (3, 0), reason="requires python3 or higher")
 @pytest.mark.settings(
     password_check_breached="strict",
     password_breached_count=16,
     password_complexity_checker="zxcvbn",
 )
 def test_breached_cnt(app, sqlalchemy_datastore):
-
     # partial response from: https://api.pwnedpasswords.com/range/07003
     pwned_response = b"AF5A73CD3CBCFDCD12B0B68CB7930F3E888:2\r\n\
 AFD8AA47E6FD782ADDC11D89744769F7354:2\r\n\
@@ -839,9 +982,7 @@ def test_authn_freshness(
     with capture_flashes() as flashes:
         response = client.get("/myspecialview", follow_redirects=False)
         assert response.status_code == 302
-        assert (
-            "/verify?next=http%3A%2F%2Flocalhost%2Fmyspecialview" in response.location
-        )
+        assert "/verify?next=http://localhost/myspecialview" in response.location
     assert flashes[0]["category"] == "error"
     assert flashes[0]["message"].encode("utf-8") == get_message(
         "REAUTHENTICATION_REQUIRED"
@@ -921,7 +1062,7 @@ def test_default_authn_bp(app, client):
     reset_fresh(client, within=timedelta(minutes=1))
     response = client.get("/myview", follow_redirects=False)
     assert response.status_code == 302
-    assert "/myprefix/verify?next=http%3A%2F%2Flocalhost%2Fmyview" in response.location
+    assert "/myprefix/verify?next=http://localhost/myview" in response.location
 
 
 def test_authn_freshness_grace(app, client, get_message):
@@ -963,7 +1104,7 @@ def test_authn_freshness_nc(app, client_nc, get_message):
     # This should fail - should be a redirect
     response = client_nc.get("/myview", headers=h, follow_redirects=False)
     assert response.status_code == 302
-    assert "/verify?next=http%3A%2F%2Flocalhost%2Fmyview" in response.location
+    assert "/verify?next=http://localhost/myview" in response.location
 
 
 def test_verify_fresh(app, client, get_message):
@@ -996,7 +1137,7 @@ def test_verify_fresh(app, client, get_message):
     response = client.post(
         verify_url, data=dict(password="password"), follow_redirects=False
     )
-    assert response.location == "http://localhost/fresh"
+    assert check_location(app, response.location, "/fresh")
 
     # should be fine now
     response = client.get("/fresh", follow_redirects=True)
@@ -1072,14 +1213,6 @@ def test_verify_next(app, client, get_message):
     )
     assert response.location == "http://localhost/mynext"
 
-    response = client.post(
-        "/auth/?next=http%3A%2F%2F127.0.0.1%3A5000%2Fdashboard%2Fsettings%2F",
-        data=dict(password="password"),
-        follow_redirects=False,
-        base_url="http://127.0.0.1:5000",
-    )
-    assert response.location == "http://127.0.0.1:5000/dashboard/settings/"
-
 
 @pytest.mark.webauthn()
 @pytest.mark.settings(webauthn_util_cls=HackWebauthnUtil)
@@ -1173,7 +1306,7 @@ def test_post_security_with_application_root_and_views(app, sqlalchemy_datastore
             "APPLICATION_ROOT": "/root",
             "SECURITY_POST_LOGIN_VIEW": "/post_login",
             "SECURITY_POST_LOGOUT_VIEW": "/post_logout",
-        }
+        },
     )
     client = app.test_client()
 
@@ -1188,6 +1321,27 @@ def test_post_security_with_application_root_and_views(app, sqlalchemy_datastore
     assert "/post_logout" in response.location
 
 
+def test_invalid_redirect_config(app, sqlalchemy_datastore, get_message):
+    with pytest.raises(ValueError):
+        init_app_with_options(
+            app,
+            sqlalchemy_datastore,
+            **{"SECURITY_REDIRECT_VALIDATE_MODE": ["regex", "bogus"]},
+        )
+
+
+def test_invalid_redirect_re(app, sqlalchemy_datastore, get_message):
+    with pytest.raises(ValueError):
+        init_app_with_options(
+            app,
+            sqlalchemy_datastore,
+            **{
+                "SECURITY_REDIRECT_VALIDATE_MODE": ["regex"],
+                "SECURITY_REDIRECT_VALIDATE_RE": None,
+            },
+        )
+
+
 @pytest.mark.settings(redirect_validate_mode="regex")
 def test_validate_redirect(app, sqlalchemy_datastore):
     """
@@ -1201,9 +1355,24 @@ def test_validate_redirect(app, sqlalchemy_datastore):
         assert not validate_redirect_url("\\\\\\github.com")
         assert not validate_redirect_url(" //github.com")
         assert not validate_redirect_url("\t//github.com")
+        assert not validate_redirect_url(r"/\github.com")
+        assert not validate_redirect_url(r"\/github.com")
         assert not validate_redirect_url("//github.com")  # this is normal urlsplit
 
 
+@pytest.mark.settings(post_login_view="\\\\\\github.com")
+def test_validate_redirect_default(app, client):
+    """
+    Test various possible URLs that urlsplit() shows as relative but
+    many browsers will interpret as absolute - and thus have a
+    open-redirect vulnerability. This tests the default configuration for
+    Flask-Security, Flask, Werkzeug
+    """
+    authenticate(client)
+    response = client.get("/login", follow_redirects=False)
+    assert response.location.startswith("http://localhost")
+
+
 def test_kwargs():
     import warnings
 
@@ -1241,3 +1410,88 @@ def test_reuse_security_object(sqlalchemy_datastore):
 
     security.init_app(app)
     assert hasattr(app, "login_manager")
+
+
+@pytest.mark.settings(static_folder_url="/mystatic/fs")
+def test_static_url(app, sqlalchemy_datastore):
+    from flask_security import url_for_security
+    from flask import url_for
+
+    init_app_with_options(app, sqlalchemy_datastore)
+    with app.test_request_context("http://localhost:5001/login"):
+        static_url = url_for_security("static", filename="js/webauthn.js")
+        assert static_url == "/mystatic/fs/js/webauthn.js"
+
+        static_url = url_for(".static", filename="js/webauthn.js")
+        assert static_url == "/mystatic/fs/js/webauthn.js"
+
+
+def test_multi_app(app, sqlalchemy_datastore):
+    # test that 2 different app with 2 different FS
+    # with USERNAME_ENABLE which dynamically changes the class definition
+    app = Flask(__name__)
+    app.response_class = Response
+    app.debug = True
+    app.config["SECRET_KEY"] = "secret"
+    app.config["TESTING"] = True
+    app.config["SECURITY_USERNAME_ENABLE"] = True
+
+    security = Security(datastore=sqlalchemy_datastore)
+    security.init_app(app)
+    assert hasattr(security.forms["register_form"].cls, "username")
+    assert "username" in security.user_identity_attributes[1].keys()
+
+    app = Flask(__name__)
+    app.response_class = Response
+    app.debug = True
+    app.config["SECRET_KEY"] = "secret"
+    app.config["TESTING"] = True
+    app.config["SECURITY_USERNAME_ENABLE"] = True
+
+    security2 = Security(datastore=sqlalchemy_datastore)
+    security2.init_app(app)
+
+    assert hasattr(security2.forms["register_form"].cls, "username")
+    assert "username" in security2.user_identity_attributes[1].keys()
+
+
+@pytest.mark.registerable()
+def test_login_email_whatever(app, client, get_message):
+    # login, by default, shouldn't verify email address is deliverable..
+    # register etc can/should do that.
+    app.config["SECURITY_EMAIL_VALIDATOR_ARGS"] = {"check_deliverability": True}
+
+    # register should fail since non-deliverable TLD
+    data = dict(
+        email="dude@me.mytld",
+        password="awesome sunset",
+    )
+    response = client.post("/register", json=data)
+    assert response.status_code == 400
+    assert response.json["response"]["errors"][0].encode("utf-8") == get_message(
+        "INVALID_EMAIL_ADDRESS"
+    )
+
+    # login should work since we are just checking for identity
+    response = client.post(
+        "/login", data=dict(email="matt@lp.com", password="password")
+    )
+    assert response.status_code == 302
+    assert "/" in response.location
+
+
+@pytest.mark.skip
+def test_sqlalchemy_session_conn(request, app, tmpdir, realdburl):
+    # test harness for checking connection mgmt by logging all connections
+    from .conftest import sqlalchemy_session_setup
+
+    ds = sqlalchemy_session_setup(
+        request, app, tmpdir, realdburl, echo_pool="debug", echo="debug"
+    )
+    init_app_with_options(app, ds)
+    client = app.test_client()
+
+    client.post("/login", data=dict(email="matt@lp.com", password="password"))
+
+    client.post("/login", json=dict(noemail="matt@lp.com", password="password"))
+    time.sleep(5)
