@@ -4,7 +4,7 @@
 
     Unified signin tests
 
-    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
 """
@@ -25,8 +25,12 @@ from tests.test_utils import (
     authenticate,
     capture_flashes,
     capture_reset_password_requests,
+    check_location,
+    get_form_action,
+    is_authenticated,
     logout,
     reset_fresh,
+    setup_tf_sms,
 )
 from tests.test_webauthn import HackWebauthnUtil, reg_2_keys
 
@@ -114,12 +118,10 @@ def us_tf_authenticate(app, client, json=False, validate=True, remember=False):
                 headers={"Content-Type": "application/json"},
             )
             assert b'"code": 200' in response.data
-            return response.json["response"].get("tf_validity_token", None)
         else:
             response = client.post(
                 "/tf-validate", data=dict(code=code), follow_redirects=True
             )
-
             assert response.status_code == 200
 
 
@@ -191,22 +193,20 @@ def test_simple_signin(app, clients, get_message):
     assert get_message("INVALID_PASSWORD_CODE") in response.data
 
     # Correct code
-    assert "remember_token" not in [c.name for c in clients.cookie_jar]
-    assert "session" not in [c.name for c in clients.cookie_jar]
+    assert not clients.get_cookie("remember_token")
+    assert not clients.get_cookie("session")
     response = clients.post(
         "/us-signin",
         data=dict(identity="matt@lp.com", passcode=requests[0]["token"]),
         follow_redirects=False,
     )
-    assert "remember_token" not in [c.name for c in clients.cookie_jar]
+    assert not clients.get_cookie("remember_token")
     assert "email" in auths[0][1]
 
-    response = clients.get("/profile", follow_redirects=False)
-    assert response.status_code == 200
+    assert is_authenticated(clients, get_message)
 
     logout(clients)
-    response = clients.get("/profile", follow_redirects=False)
-    assert "/login?next=%2Fprofile" in response.location
+    assert not is_authenticated(clients, get_message)
 
     # login via SMS
     sms_sender = SmsSenderFactory.createSender("test")
@@ -226,14 +226,13 @@ def test_simple_signin(app, clients, get_message):
         follow_redirects=True,
     )
     assert response.status_code == 200
-    assert "remember_token" in [c.name for c in clients.cookie_jar]
+    assert clients.get_cookie("remember_token")
     assert "sms" in auths[1][1]
 
-    response = clients.get("/profile", follow_redirects=False)
-    assert response.status_code == 200
+    assert is_authenticated(clients, get_message)
 
     logout(clients)
-    assert "remember_token" not in [c.name for c in clients.cookie_jar]
+    assert not clients.get_cookie("remember_token")
 
 
 def test_simple_signin_json(app, client_nc, get_message):
@@ -247,7 +246,6 @@ def test_simple_signin_json(app, client_nc, get_message):
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
 
     with capture_flashes() as flashes:
-
         response = client_nc.get("/us-signin", headers=headers)
         jresponse = response.json["response"]
         assert (
@@ -293,8 +291,7 @@ def test_simple_signin_json(app, client_nc, get_message):
         assert "email" in auths[0][1]
 
         logout(client_nc)
-        response = client_nc.get("/profile", headers=headers, follow_redirects=False)
-        assert response.status_code == 401
+        assert not is_authenticated(client_nc, get_message)
 
         # login via SMS
         sms_sender = SmsSenderFactory.createSender("test")
@@ -386,7 +383,7 @@ def test_us_passwordless(app, client, get_message):
         token = matcher[2].split(":")[1]
         assert token == requests[0]["token"]  # deprecated
         assert token == requests[0]["login_token"]
-        assert matcher[3].split(":")[1] == app.config["SECURITY_CONFIRM_URL"]
+        assert matcher[3].split(":")[1] == "True"  # register_blueprint
         assert matcher[4].split(":")[1] == "nopasswd-dude@lp.com"
 
         # check link
@@ -458,9 +455,8 @@ def test_us_passwordless_confirm_json(app, client, get_message):
     outbox = app.mail.outbox
     matcher = re.findall(r"\w+:.*", outbox[0].body, re.IGNORECASE)
     link = matcher[0].split(":", 1)[1]
-    response = client.get(link, headers=headers, follow_redirects=True)
-    assert get_message("EMAIL_CONFIRMED") in response.data
-    logout(client)
+    response = client.get(link, headers=headers, follow_redirects=False)
+    assert check_location(app, response.location, "/login")
 
     # should be able to authenticate now.
     response = client.post(
@@ -584,6 +580,10 @@ def test_post_already_authenticated(client, get_message):
     assert b"Post Login" in response.data
     response = client.post("/us-signin?next=/page1", data=data, follow_redirects=True)
     assert b"Page 1" in response.data
+    # should work in form as well
+    ndata = dict(email="matt@lp.com", password="password", next="/page1")
+    response = client.post("/us-signin", data=ndata, follow_redirects=True)
+    assert b"Page 1" in response.data
 
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
     response = client.post("/us-signin", json=data, headers=headers)
@@ -643,8 +643,7 @@ def test_verify_link(app, client, get_message):
     assert "email" in auths[0][1]
 
     # verify logged in
-    response = client.get("/profile", follow_redirects=False)
-    assert response.status_code == 200
+    assert is_authenticated(client, get_message)
 
 
 @pytest.mark.settings(
@@ -711,8 +710,7 @@ def test_verify_link_spa(app, client, get_message):
     qparams = dict(parse_qsl(split.query))
     assert qparams["email"] == "matt@lp.com"
 
-    response = client.get("/profile", headers=headers)
-    assert response.status_code == 200
+    assert is_authenticated(client, get_message)
 
 
 def test_setup(app, client, get_message):
@@ -747,12 +745,8 @@ def test_setup(app, client, get_message):
     assert response.status_code == 200
     assert b"Submit Code" in response.data
     assert b"Enter code here to complete setup" in response.data
-    matcher = re.match(
-        r'.*<form action="([^\s]*)".*',
-        response.data.decode("utf-8"),
-        re.IGNORECASE | re.DOTALL,
-    )
-    verify_url = matcher.group(1)
+
+    verify_url = get_form_action(response, 1)
 
     # Try invalid code
     response = client.post(verify_url, data=dict(passcode=12345), follow_redirects=True)
@@ -774,12 +768,7 @@ def test_setup_email(app, client, get_message):
     assert b"Enter code here to complete setup" in response.data
     outbox = app.mail.outbox
 
-    matcher = re.match(
-        r'.*<form action="([^\s]*)".*',
-        response.data.decode("utf-8"),
-        re.IGNORECASE | re.DOTALL,
-    )
-    verify_url = matcher.group(1)
+    verify_url = get_form_action(response, 1)
 
     # verify no magic link - us_authenticate received first email - we want the second
     matcher = re.findall(r"\w+:.*", outbox[1].body, re.IGNORECASE)
@@ -998,17 +987,16 @@ def test_verify(app, client, get_message):
     us_authenticate(client)
     response = client.get("us-setup", follow_redirects=False)
     verify_url = response.location
-    assert "/us-verify?next=http%3A%2F%2Flocalhost%2Fus-setup" in verify_url
+    assert check_location(
+        app, response.location, "/us-verify?next=http://localhost/us-setup"
+    )
     logout(client)
     us_authenticate(client)
 
     response = client.get("us-setup", follow_redirects=True)
     form_response = response.data.decode("utf-8")
     assert "Please Reauthenticate" in form_response
-    matcher = re.match(
-        r'.*formaction="([^"]*)".*', form_response, re.IGNORECASE | re.DOTALL
-    )
-    send_code_url = matcher.group(1)
+    send_code_url = get_form_action(response, 1)
 
     # Send unknown method
     response = client.post(
@@ -1032,7 +1020,7 @@ def test_verify(app, client, get_message):
 
     code = sms_sender.messages[0].split()[-1].strip(".")
     response = client.post(verify_url, data=dict(passcode=code), follow_redirects=False)
-    assert response.location == "http://localhost/us-setup"
+    assert check_location(app, response.location, "/us-setup")
 
 
 def test_verify_json(app, client, get_message):
@@ -1237,12 +1225,7 @@ def test_qrcode(app, client, get_message):
     code = totp.generate().token
 
     # get verify link e.g. /us-setup/{state}
-    matcher = re.match(
-        r'.*<form action="([^\s]*)".*',
-        response.data.decode("utf-8"),
-        re.IGNORECASE | re.DOTALL,
-    )
-    verify_url = matcher.group(1)
+    verify_url = get_form_action(response, 1)
 
     response = client.post(verify_url, data=dict(passcode=code), follow_redirects=True)
     assert response.status_code == 200
@@ -1267,6 +1250,7 @@ def test_next(app, client, get_message):
     assert "/post_login" in response.location
 
     logout(client)
+    # Test form.next
     response = client.post(
         "/us-signin",
         data=dict(
@@ -1324,8 +1308,7 @@ def test_confirmable(app, client, get_message):
     assert get_message("CONFIRMATION_REQUIRED") in response.data
 
     # Verify not authenticated
-    response = client.get("/profile", follow_redirects=False)
-    assert "/login?next=%2Fprofile" in response.location
+    assert not is_authenticated(client, get_message)
 
 
 @pytest.mark.registerable()
@@ -1348,7 +1331,7 @@ def test_can_add_password(app, client, get_message):
         follow_redirects=True,
     )
 
-    assert get_message("PASSWORD_RESET") in response.data
+    assert get_message("PASSWORD_RESET_NO_LOGIN") in response.data
 
     # authenticate with new password using standard/old login endpoint.
     response = authenticate(
@@ -1457,11 +1440,9 @@ def test_regular_login(app, client, get_message):
         follow_redirects=True,
     )
     assert response.status_code == 200
-    assert "remember_token" in [c.name for c in client.cookie_jar]
+    assert client.get_cookie("remember_token")
     assert "password" in auths[0][1]
-
-    response = client.get("/profile", follow_redirects=False)
-    assert response.status_code == 200
+    assert is_authenticated(client, get_message)
 
 
 @pytest.mark.settings(
@@ -1581,12 +1562,7 @@ def test_tf_not(app, client, get_message):
     response = client.post(
         "us-setup", data=dict(chosen_method="sms", phone="650-555-1212")
     )
-    matcher = re.match(
-        r'.*<form action="([^\s]*)".*',
-        response.data.decode("utf-8"),
-        re.IGNORECASE | re.DOTALL,
-    )
-    verify_url = matcher.group(1)
+    verify_url = get_form_action(response, 1)
     code = sms_sender.messages[0].split()[-1].strip(".")
     response = client.post(verify_url, data=dict(passcode=code), follow_redirects=True)
     assert response.status_code == 200
@@ -1612,8 +1588,7 @@ def test_tf_not(app, client, get_message):
     # assert "sms" in auths[1][1]
 
     # Verify authenticated
-    response = client.get("/profile", follow_redirects=False)
-    assert response.status_code == 200
+    assert is_authenticated(client, get_message)
 
 
 @pytest.mark.settings(sms_service="bad")
@@ -1834,45 +1809,22 @@ def test_totp_generation(app, client, get_message):
 )
 def test_us_tf_validity(app, client, get_message):
     us_tf_authenticate(app, client, remember=True)
+    assert client.get_cookie("tf_validity")
     logout(client)
-    data = dict(identity="gal@lp.com", passcode="password")
-    response = client.post(
-        "/us-signin", json=data, headers={"Content-Type": "application/json"}
-    )
-    assert b'"code": 200' in response.data
-    cookie = next(
-        (cookie for cookie in client.cookie_jar if cookie.name == "tf_validity"), None
-    )
-    assert cookie is not None
+    # logout does NOT remove this cookie
+    assert client.get_cookie("tf_validity")
 
+    # This time shouldn't require code
+    data = dict(identity="gal@lp.com", passcode="password")
+    response = client.post("/us-signin", json=data)
+    assert response.json["meta"]["code"] == 200
+    assert is_authenticated(client, get_message)
     logout(client)
 
     data = dict(identity="gal2@lp.com", passcode="password")
     response = client.post("/us-signin", data=data, follow_redirects=True)
     assert b"Please enter your authentication code" in response.data
 
-    # clear the cookie to make sure it's not picking it up with json.
-    client.cookie_jar.clear("localhost.local", "/", "tf_validity")
-
-    token = us_tf_authenticate(app, client, remember=True, json=True)
-    logout(client)
-    data = dict(identity="gal@lp.com", passcode="password", tf_validity_token=token)
-    response = client.post(
-        "/us-signin",
-        json=data,
-        follow_redirects=True,
-        headers={"Content-Type": "application/json"},
-    )
-    assert response.status_code == 200
-    # verify logged in
-    response = client.get("/profile", follow_redirects=False)
-
-    assert response.status_code == 200
-    assert b"Welcome gal@lp.com" in response.data
-
-    logout(client)
-
-    data["identity"] = "gal2@lp.com"
     response = client.post(
         "/us-signin",
         json=data,
@@ -2032,7 +1984,6 @@ def test_setup_delete_multi_json(app, client, get_message):
 
 @pytest.mark.settings(return_generic_responses=True)
 def test_generic_response(app, client, get_message):
-
     # test not-setup choice
     data = dict(identity="matt@lp.com", chosen_method="email")
     response = client.post("/us-signin/send-code", data=data, follow_redirects=True)
@@ -2117,3 +2068,64 @@ def test_generic_response(app, client, get_message):
         "us-verify-link?email=matt@lp.com&code=12345", follow_redirects=True
     )
     assert get_message("GENERIC_AUTHN_FAILED") in response.data
+
+
+@pytest.mark.settings(url_prefix="/auth", us_signin_replaces_login=True)
+def test_propagate_next(app, client):
+    # verify we propagate the ?next param all the way through a unified signin
+    # Also test blueprint prefix since we rarely actually test that.
+    set_phone(app)
+    with capture_send_code_requests() as codes:
+        response = client.get("profile", follow_redirects=True)
+        assert "?next=/profile" in response.request.url
+        signin_url = get_form_action(response, 0)
+        sendcode_url = get_form_action(response, 1)
+        response = client.post(
+            sendcode_url, data=dict(identity="matt@lp.com", chosen_method="sms")
+        )
+        data = dict(identity="matt@lp.com", passcode=codes[0]["login_token"])
+        response = client.post(signin_url, data=data, follow_redirects=False)
+        assert "/profile" in response.location
+
+
+@pytest.mark.two_factor()
+@pytest.mark.settings(url_prefix="/auth", us_signin_replaces_login=True)
+def test_propagate_next_tf(app, client):
+    # test next is propagated with a second factor
+    response = client.post(
+        "/auth/login", json=dict(identity="matt@lp.com", passcode="password")
+    )
+    sms_sender = setup_tf_sms(client, url_prefix=app.config["SECURITY_URL_PREFIX"])
+    logout(client, endpoint="/auth/logout")
+
+    response = client.get("/profile", follow_redirects=True)
+    assert "?next=/profile" in response.request.url
+    signin_url = get_form_action(response, 0)
+    response = client.post(
+        signin_url,
+        data=dict(identity="matt@lp.com", passcode="password"),
+        follow_redirects=True,
+    )
+    sendcode_url = get_form_action(response, 0)
+    response = client.post(
+        sendcode_url,
+        data=dict(code=sms_sender.messages[0].split()[-1]),
+        follow_redirects=True,
+    )
+    assert b"Profile Page" in response.data
+
+    # Try form.next
+    logout(client, endpoint="/auth/logout")
+
+    response = client.post(
+        "/auth/us-signin",
+        data=dict(identity="matt@lp.com", passcode="password", next="/im-in"),
+        follow_redirects=True,
+    )
+    sendcode_url = get_form_action(response, 0)
+    response = client.post(
+        sendcode_url,
+        data=dict(code=sms_sender.messages[0].split()[-1]),
+        follow_redirects=False,
+    )
+    assert "/im-in" in response.location
