@@ -1,10 +1,10 @@
 """
-    utils
-    ~~~~~
+    test_utils
+    ~~~~~~~~~~
 
     Test utils
 
-    :copyright: (c) 2019-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2019-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 """
 from contextlib import contextmanager
@@ -14,7 +14,7 @@ import typing as t
 from flask.json.tag import TaggedJSONSerializer
 from flask.signals import message_flashed
 
-from flask_security import Security, SmsSenderBaseClass
+from flask_security import Security, SmsSenderBaseClass, SmsSenderFactory, UserMixin
 from flask_security.datastore import (
     SQLAlchemyUserDatastore,
     SQLAlchemySessionUserDatastore,
@@ -22,9 +22,12 @@ from flask_security.datastore import (
 from flask_security.signals import (
     login_instructions_sent,
     reset_password_instructions_sent,
+    tf_security_token_sent,
     user_registered,
+    us_security_token_sent,
 )
 from flask_security.utils import hash_data, hash_password
+from flask_security.utils import config_value as cv
 
 from itsdangerous import BadSignature, SignatureExpired, URLSafeTimedSerializer
 from werkzeug.http import parse_cookie
@@ -40,11 +43,35 @@ def authenticate(
 
 
 def json_authenticate(client, email="matt@lp.com", password="password", endpoint=None):
-    data = f'{{"email": "{email}", "password": "{password}"}}'
+    data = dict(email=email, password=password)
 
     # Get auth token always
     ep = endpoint or "/login?include_auth_token"
-    return client.post(ep, content_type="application/json", data=data)
+    return client.post(ep, content_type="application/json", json=data)
+
+
+def is_authenticated(client, get_message):
+    # Return True is 'client' is authenticated.
+    # Return False is not
+    # Raise ValueError not certain...
+    response = client.get("/profile", headers={"accept": "application/json"})
+    if response.status_code == 200:
+        return True
+    if response.status_code == 401 and response.json["response"]["errors"][0].encode(
+        "utf-8"
+    ) == get_message("UNAUTHENTICATED"):
+        return False
+    raise ValueError("Failed to figure out if authenticated")
+
+
+def check_location(app, location, expected_base):
+    # verify response location. This can be absolute or relative based
+    # on configuration
+    redirect_validate_mode = cv("REDIRECT_VALIDATE_MODE", app=app) or []
+    if "absolute" in redirect_validate_mode:
+        return location == f"http://localhost{expected_base}"
+    else:
+        return location == expected_base
 
 
 def verify_token(client_nc, token, status=None):
@@ -75,6 +102,15 @@ def json_logout(client, token, endpoint=None):
     )
 
 
+def get_csrf_token(client):
+    response = client.get(
+        "/login",
+        data={},
+        headers={"Accept": "application/json"},
+    )
+    return response.json["response"]["csrf_token"]
+
+
 def get_session(response):
     """Return session cookie contents.
     This a base64 encoded json.
@@ -94,10 +130,24 @@ def get_session(response):
                 return val[1]
 
 
-def get_existing_session(client):
-    cookie = next(
-        (cookie for cookie in client.cookie_jar if cookie.name == "session"), None
+def setup_tf_sms(client, url_prefix=None):
+    # Simple setup of SMS as a second factor and return the sender so caller
+    # can get codes.
+    SmsSenderFactory.senders["test"] = SmsTestSender
+    sms_sender = SmsSenderFactory.createSender("test")
+    data = dict(setup="sms", phone="+442083661188")
+    response = client.post("/".join(filter(None, (url_prefix, "tf-setup"))), json=data)
+    assert sms_sender.get_count() == 1
+    code = sms_sender.messages[0].split()[-1]
+    response = client.post(
+        "/".join(filter(None, (url_prefix, "tf-validate"))), json=dict(code=code)
     )
+    assert response.status_code == 200
+    return sms_sender
+
+
+def get_existing_session(client):
+    cookie = client.get_cookie("session")
     if cookie:
         serializer = URLSafeTimedSerializer("secret", serializer=TaggedJSONSerializer())
         val = serializer.loads_unsafe(cookie.value)
@@ -115,6 +165,30 @@ def reset_fresh(client, within):
     return old_paa
 
 
+def get_form_action(response, ordinal=0):
+    # Return the URL that the form WOULD post to - this is useful to check
+    # how our templates actually work (e.g. propagation of 'next')
+    matcher = re.findall(
+        r'(?:<form action|formaction)="(\S*)"',
+        response.data.decode("utf-8"),
+        re.IGNORECASE | re.DOTALL,
+    )
+    return matcher[ordinal]
+
+
+def get_form_input(response, field_id):
+    # return value of field with the id == field_id or None if not found
+    rex = f'<input id="{field_id}"[^>]*value="([^"]*)">'
+    matcher = re.findall(
+        rex,
+        response.data.decode("utf-8"),
+        re.IGNORECASE | re.DOTALL,
+    )
+    if matcher:
+        return matcher[0]
+    return None
+
+
 def check_xlation(app, locale):
     """Return True if locale is loaded"""
     with app.test_request_context():
@@ -201,9 +275,9 @@ def get_num_queries(datastore):
     return None if datastore doesn't support this.
     """
     if is_sqlalchemy(datastore):
-        from flask_sqlalchemy import get_debug_queries
+        from flask_sqlalchemy.record_queries import get_recorded_queries
 
-        return len(get_debug_queries())
+        return len(get_recorded_queries())
     return None
 
 
@@ -306,6 +380,27 @@ def capture_flashes():
         message_flashed.disconnect(_on)
 
 
+@contextmanager
+def capture_send_code_requests():
+    # Easy way to get token/code required for code logins
+    # either second factor or us_signin
+    login_requests = []
+
+    def _on(app, **data):
+        assert all(v in data for v in ["user", "method", "login_token"])
+        assert isinstance(data["user"], UserMixin)
+        login_requests.append(data)
+
+    us_security_token_sent.connect(_on)
+    tf_security_token_sent.connect(_on)
+
+    try:
+        yield login_requests
+    finally:
+        us_security_token_sent.disconnect(_on)
+        tf_security_token_sent.disconnect(_on)
+
+
 def get_auth_token_version_3x(app, user):
     """
     Copy of algorithm that generated user token in version 3.x
