@@ -4,7 +4,7 @@
 
     WebAuthn tests
 
-    :copyright: (c) 2021-2022 by J. Christopher Wagner (jwag).
+    :copyright: (c) 2021-2023 by J. Christopher Wagner (jwag).
     :license: MIT, see LICENSE for more details.
 
 """
@@ -22,17 +22,18 @@ import pytest
 from tests.test_two_factor import tf_in_session
 from tests.test_utils import (
     FakeSerializer,
-    SmsTestSender,
     authenticate,
     capture_flashes,
     get_existing_session,
+    get_form_action,
+    get_form_input,
     json_authenticate,
     logout,
     reset_fresh,
+    setup_tf_sms,
 )
 
 from flask_security import (
-    SmsSenderFactory,
     WebauthnUtil,
     user_authenticated,
     wan_registered,
@@ -60,9 +61,9 @@ REG_DATA1 = {
         "FEybDVYMnN5UTNGUmVXUlRVVjlyVUVWcVZqVmhNbVF3UVhCbVlYUmpjRk"
         "V4WVZoRWJWRlFidyIsIm9yaWdpbiI6Imh0dHA6Ly9sb2NhbGhvc3Q6NT"
         "AwMSIsImNyb3NzT3JpZ2luIjpmYWxzZX0",
+        "transports": ["usb"],
     },
     "extensions": '{"credProps": {}}',
-    "transports": ["usb"],
 }
 SIGNIN_DATA1 = {
     "id": "wUUqNOjY35dcT-vpikZpZx-T91NjIe4PqrV8j7jYPOc",
@@ -94,9 +95,9 @@ REG_DATA2 = {
         "YzIxRFEybDVYMnN5UTNGUmVXUlRVVjlyVUVWcVZqVmhNbVF3UV"
         "hCbVlYUmpjRkV4WVZoRWJWRlFidyIsIm9yaWdpbiI6Imh0dHA6L"
         "y9sb2NhbGhvc3Q6NTAwMSIsImNyb3NzT3JpZ2luIjpmYWxzZX0",
+        "transports": ["nfc", "usb"],
     },
     "extensions": '{"credProps": {"rk": True}}',
-    "transports": ["nfc", "usb"],
 }
 
 # This has user_verification=True - i.e. a multi-factor capable key
@@ -115,9 +116,9 @@ REG_DATA_UV = {
         "xRFEybDVYMnN5UTNGUmVXUlRVVjlyVUVWcVZqVmhNbVF3UVhCbVlY"
         "UmpjRkV4WVZoRWJWRlFidyIsIm9yaWdpbiI6Imh0dHA6Ly9sb2Nhb"
         "Ghvc3Q6NTAwMSIsImNyb3NzT3JpZ2luIjpmYWxzZX0",
+        "transports": ["nfc", "usb"],
     },
     "extensions": '{"credProps":{"rk":true}}',
-    "transports": ["nfc", "usb"],
 }
 
 SIGNIN_DATA_UV = {
@@ -154,9 +155,9 @@ REG_DATA_UH = {
         "yVUVWcVZqVmhNbVF3UVhCbVlYUmpjRkV4WVZoRWJWRlFidyIsIm9yaWdpbi"
         "I6Imh0dHA6Ly9sb2NhbGhvc3"
         "Q6NTAwMSIsImNyb3NzT3JpZ2luIjpmYWxzZX0",
+        "transports": ["nfc", "usb"],
     },
     "extensions": '{"credProps":{"rk": true}}"',
-    "transports": ["nfc", "usb"],
 }
 SIGNIN_DATA_UH = {
     "id": "rHb1OXVM--dgGcWg0u3cfomyc-Tu4l4kK8GjVkS8bms-foXmBAlWHyTzuhgGgCnx",
@@ -185,35 +186,15 @@ class HackWebauthnUtil(WebauthnUtil):
         return "http://localhost:5001"
 
 
-def setup_tf(client):
-    sms_sender = SmsSenderFactory.createSender("test")
-    data = dict(setup="sms", phone="+442083661188")
-    client.post("/tf-setup", json=data)
-    assert sms_sender.get_count() == 1
-    code = sms_sender.messages[0].split()[-1]
-    response = client.post("/tf-validate", json=dict(code=code))
-    assert response.status_code == 200
-    return sms_sender
-
-
-SmsSenderFactory.senders["test"] = SmsTestSender
-
-
-def _register_start(client, name="testr1", usage="secondary"):
-    response = client.post("wan-register", data=dict(name=name, usage=usage))
+def _register_start(client, name="testr1", usage="secondary", endpoint="wan-register"):
+    response = client.post(endpoint, data=dict(name=name, usage=usage))
     matcher = re.match(
         r".*handleRegister\(\'(.*)\'\).*",
         response.data.decode("utf-8"),
         re.IGNORECASE | re.DOTALL,
     )
     register_options = json.loads(matcher.group(1))
-
-    action_matcher = re.match(
-        r'.*<form action="([^\s]*)".*',
-        response.data.decode("utf-8"),
-        re.IGNORECASE | re.DOTALL,
-    )
-    response_url = action_matcher.group(1)
+    response_url = get_form_action(response)
     return register_options, response_url
 
 
@@ -258,33 +239,30 @@ def _signin_start(
         re.IGNORECASE | re.DOTALL,
     )
     signin_options = json.loads(matcher.group(1))
-
-    action_matcher = re.match(
-        r'.*<form action="([^\s]*)".*',
-        response.data.decode("utf-8"),
-        re.IGNORECASE | re.DOTALL,
-    )
-    response_url = action_matcher.group(1)
+    response_url = get_form_action(response)
     return signin_options, response_url
 
 
-def _signin_start_json(client, identity=None, remember=False):
+def _signin_start_json(client, identity=None, remember=False, endpoint="wan-signin"):
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
 
     response = client.post(
-        "wan-signin", headers=headers, json=dict(identity=identity, remember=remember)
+        endpoint, headers=headers, json=dict(identity=identity, remember=remember)
     )
     signin_options = response.json["response"]["credential_options"]
     response_url = f'wan-signin/{response.json["response"]["wan_state"]}'
     return signin_options, response_url, response.json
 
 
-def wan_signin(client, identity, signin_data):
+def wan_signin(client, identity, signin_data, wan_signin_url):
     # perform complete sign in - useful for tests outside this module.
-    signin_options, response_url, _ = _signin_start_json(client, identity)
+    signin_options, response_url = _signin_start(
+        client, identity, endpoint=wan_signin_url
+    )
     response = client.post(
         response_url,
-        json=dict(credential=json.dumps(signin_data)),
+        data=dict(credential=json.dumps(signin_data)),
+        follow_redirects=True,
     )
     assert response.status_code == 200
     return response
@@ -397,7 +375,9 @@ def test_basic_json(app, clients, get_message):
             cred.lastuse_datetime = fake_dt
             app.security.datastore.put(cred)
         app.security.datastore.commit()
-        if hasattr(app.security.datastore.db, "close_db"):
+        if hasattr(app.security.datastore.db, "close_db") and callable(
+            app.security.datastore.db.close_db
+        ):
             app.security.datastore.db.close_db(None)
 
     response = clients.get("/wan-register", headers=headers)
@@ -796,7 +776,7 @@ def test_tf(app, client, get_message):
     assert response.status_code == 200
     assert b"Use Your WebAuthn Security Key as a Second Factor" in response.data
     # we should have a wan key available
-    assert b'action="/wan-signin"' in response.data
+    assert b'action="/wan-signin' in response.data
 
     # verify NOT logged in
     response = client.get("/profile", follow_redirects=False)
@@ -862,7 +842,7 @@ def test_tf_json(app, client, get_message):
 def test_tf_validity_window(app, client, get_message):
     # Test with a two-factor validity setting - we don't get re-prompted.
     authenticate(client)
-    assert "tf_validity" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("tf_validity")
     register_options, response_url = _register_start_json(client)
     client.post(response_url, json=dict(credential=json.dumps(REG_DATA1)))
     logout(client)
@@ -898,7 +878,7 @@ def test_tf_validity_window(app, client, get_message):
     signin_options, response_url = _signin_start(client, "matt@lp.com")
     response = client.post(response_url, json=dict(credential=json.dumps(SIGNIN_DATA1)))
     assert response.status_code == 200
-    assert "tf_validity" in [c.name for c in client.cookie_jar]
+    assert client.get_cookie("tf_validity")
     logout(client)
 
     # since we did specify 'remember' previously - should not require 2FA
@@ -921,7 +901,8 @@ def test_tf_validity_window(app, client, get_message):
 )
 def test_tf_validity_window_json(app, client, get_message):
     # Test with a two-factor validity setting - we don't get re-prompted.
-    response = json_authenticate(client)
+    # This also relies on the tf_validity_cookie
+    json_authenticate(client)
     register_options, response_url = _register_start_json(client)
     client.post(response_url, json=dict(credential=json.dumps(REG_DATA1)))
     logout(client)
@@ -935,12 +916,8 @@ def test_tf_validity_window_json(app, client, get_message):
     signin_options, response_url = _signin_start(client, "matt@lp.com")
     response = client.post(response_url, json=dict(credential=json.dumps(SIGNIN_DATA1)))
     assert response.status_code == 200
-    tf_token = response.json["response"]["tf_validity_token"]
     logout(client)
 
-    # make sure the cookie doesn't affect the JSON request
-    client.cookie_jar.clear("localhost.local", "/", "tf_validity")
-
     # Sign in again - shouldn't require 2FA
     response = client.post(
         "/login",
@@ -948,7 +925,6 @@ def test_tf_validity_window_json(app, client, get_message):
             email="matt@lp.com",
             password="password",
             remember=True,
-            tf_validity_token=tf_token,
         ),
     )
     assert response.status_code == 200
@@ -960,7 +936,6 @@ def test_tf_validity_window_json(app, client, get_message):
     webauthn_util_cls=HackWebauthnUtil, wan_register_within="1 seconds"
 )
 def test_register_timeout(app, client, get_message):
-
     authenticate(client)
 
     app.security.wan_serializer = FakeSerializer(1.0)
@@ -975,7 +950,6 @@ def test_register_timeout(app, client, get_message):
 
 @pytest.mark.settings(webauthn_util_cls=HackWebauthnUtil, wan_signin_within="2 seconds")
 def test_signin_timeout(app, client, get_message):
-
     authenticate(client)
 
     register_options, response_url = _register_start_json(client, name="testr3")
@@ -1116,7 +1090,7 @@ def test_alt_tf(app, client, get_message):
     response = client.post(response_url, json=dict(credential=json.dumps(REG_DATA1)))
     assert response.status_code == 200
 
-    sms_sender = setup_tf(client)
+    sms_sender = setup_tf_sms(client)
     logout(client)
 
     # sign in using webauthn key
@@ -1145,7 +1119,7 @@ def test_all_in_one(app, client, get_message):
     register_options, response_url = _register_start_json(client, usage="first")
     response = client.post(response_url, json=dict(credential=json.dumps(REG_DATA_UV)))
     assert response.status_code == 200
-    setup_tf(client)
+    setup_tf_sms(client)
 
     logout(client)
     signin_options, response_url, rjson = _signin_start_json(client, "matt@lp.com")
@@ -1168,7 +1142,7 @@ def test_all_in_one_not_allowed(app, client, get_message):
     register_options, response_url = _register_start_json(client, usage="first")
     response = client.post(response_url, json=dict(credential=json.dumps(REG_DATA_UV)))
     assert response.status_code == 200
-    setup_tf(client)
+    setup_tf_sms(client)
     logout(client)
 
     app.config["SECURITY_WAN_ALLOW_AS_MULTI_FACTOR"] = False
@@ -1204,20 +1178,20 @@ def test_reset(app, client):
 
 
 @pytest.mark.settings(webauthn_util_cls=HackWebauthnUtil)
-def test_user_handle(app, client, get_message):
+def test_user_handle(app, clients, get_message):
     """Test that we fail signin if user_handle doesn't match.
     Since we generated the SIGNIN_DATA_OH from view_scaffold - the user_handle
     has no way of matching.
     """
-    authenticate(client)
-    register_options, response_url = _register_start_json(client, usage="first")
-    response = client.post(response_url, json=dict(credential=json.dumps(REG_DATA_UH)))
+    authenticate(clients)
+    register_options, response_url = _register_start_json(clients, usage="first")
+    response = clients.post(response_url, json=dict(credential=json.dumps(REG_DATA_UH)))
     assert response.status_code == 200
 
     # verify can't sign in
-    logout(client)
-    signin_options, response_url, _ = _signin_start_json(client, "matt@lp.com")
-    response = client.post(
+    logout(clients)
+    signin_options, response_url, _ = _signin_start_json(clients, "matt@lp.com")
+    response = clients.post(
         response_url, json=dict(credential=json.dumps(SIGNIN_DATA_UH))
     )
     assert response.json["response"]["field_errors"]["credential"][0].encode(
@@ -1235,15 +1209,15 @@ def test_user_handle(app, client, get_message):
             .decode("utf-8")
             .replace("=", "")
         )
-        upd_signin_data = copy.deepcopy(SIGNIN_DATA_UH)
-        upd_signin_data["response"]["userHandle"] = b64_user_handle
-        signin_options, response_url, _ = _signin_start_json(client, "matt@lp.com")
-        response = client.post(
-            response_url, json=dict(credential=json.dumps(upd_signin_data))
-        )
-        # verify actually logged in
-        response = client.get("/profile", headers={"accept": "application/json"})
-        assert response.status_code == 200
+    upd_signin_data = copy.deepcopy(SIGNIN_DATA_UH)
+    upd_signin_data["response"]["userHandle"] = b64_user_handle
+    signin_options, response_url, _ = _signin_start_json(clients, "matt@lp.com")
+    response = clients.post(
+        response_url, json=dict(credential=json.dumps(upd_signin_data))
+    )
+    # verify actually logged in
+    response = clients.get("/profile", headers={"accept": "application/json"})
+    assert response.status_code == 200
 
 
 @pytest.mark.settings(webauthn_util_cls=HackWebauthnUtil)
@@ -1318,7 +1292,7 @@ def test_verify(app, client, get_message):
 
     old_paa = reset_fresh(client, app.config["SECURITY_FRESHNESS"])
     response = client.get("fresh")
-    assert "/verify?next=http%3A%2F%2Flocalhost%2Ffresh" in response.location
+    assert "/verify?next=http://localhost/fresh" in response.location
     signin_options, response_url = _signin_start(
         client, endpoint="wan-verify?next=/fresh"
     )
@@ -1488,7 +1462,7 @@ def test_remember_token(client):
     assert response.status_code == 200
     logout(client)
 
-    assert "remember_token" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("remember_token")
 
     headers = {"Accept": "application/json", "Content-Type": "application/json"}
 
@@ -1502,8 +1476,8 @@ def test_remember_token(client):
         response_url,
         json=dict(credential=json.dumps(SIGNIN_DATA1), remember=True),
     )
-    assert "remember_token" in [c.name for c in client.cookie_jar]
-    client.cookie_jar.clear_session_cookies()
+    assert client.get_cookie("remember_token")
+    client.delete_cookie("session")
     response = client.get("/profile")
     assert b"profile" in response.data
 
@@ -1519,7 +1493,7 @@ def test_remember_token_tf(client):
     assert response.status_code == 200
     logout(client)
 
-    assert "remember_token" not in [c.name for c in client.cookie_jar]
+    assert not client.get_cookie("remember_token")
 
     # login again - should require MFA
     response = client.post(
@@ -1537,8 +1511,8 @@ def test_remember_token_tf(client):
         response_url,
         json=dict(credential=json.dumps(SIGNIN_DATA1), remember=True),
     )
-    assert "remember_token" in [c.name for c in client.cookie_jar]
-    client.cookie_jar.clear_session_cookies()
+    assert client.get_cookie("remember_token")
+    client.delete_cookie("session")
     response = client.get("/profile")
     assert b"profile" in response.data
 
@@ -1648,3 +1622,53 @@ def test_mf(client):
     # verify actually logged in
     response = client.get("/profile", follow_redirects=False)
     assert response.status_code == 200
+
+
+@pytest.mark.settings(webauthn_util_cls=HackWebauthnUtil, url_prefix="/auth")
+def test_login_next(app, client, get_message):
+    # Test that ?next=/xx is propagated through login/wan-signin templates as well as
+    # views.
+    # Also - use a different blueprint prefix - we rarely test that....
+    authenticate(client, endpoint="/auth/login")
+    register_options, response_url = _register_start(
+        client, name="testr3", usage="first", endpoint="/auth/wan-register"
+    )
+    response = client.post(
+        response_url, data=dict(credential=json.dumps(REG_DATA1)), follow_redirects=True
+    )
+    assert response.status_code == 200
+    assert get_message("WEBAUTHN_REGISTER_SUCCESSFUL", name="testr3") in response.data
+    logout(client, endpoint="/auth/logout")
+
+    response = client.get("profile", follow_redirects=True)
+    assert "?next=/profile" in response.request.url
+    # pull webauthn form action out of login_form - should have ?next=...
+    webauthn_url = get_form_action(response, 1)
+
+    signin_options, response_url = _signin_start(
+        client, "matt@lp.com", endpoint=webauthn_url
+    )
+    response = client.post(
+        response_url,
+        data=dict(credential=json.dumps(SIGNIN_DATA1)),
+        follow_redirects=True,
+    )
+    assert response.status_code == 200
+    assert b"Profile Page" in response.data
+
+    # Try form.next
+    logout(client, endpoint="/auth/logout")
+    reset_signcount(app, "matt@lp.com", "testr3")
+
+    response = client.post(
+        "/auth/wan-signin", data=dict(identity="matt@lp.com", next="/im-in")
+    )
+    response_url = get_form_action(response)
+
+    next_loc = get_form_input(response, "next")
+    response = client.post(
+        response_url,
+        data=dict(credential=json.dumps(SIGNIN_DATA1), next=next_loc),
+        follow_redirects=False,
+    )
+    assert "/im-in" in response.location
